@@ -1,4 +1,5 @@
 package HTML::LinkList;
+$HTML::LinkList::VERSION = '0.1701';
 use strict;
 use warnings;
 
@@ -8,11 +9,7 @@ HTML::LinkList - Create a 'smart' list of HTML links.
 
 =head1 VERSION
 
-This describes version B<0.1501> of HTML::LinkList.
-
-=cut
-
-our $VERSION = '0.1501';
+version 0.1701
 
 =head1 SYNOPSIS
 
@@ -225,6 +222,10 @@ String to append to each item.
 An additional string to put in front of each "active" item, after pre_item.
 The "active" item is the link which matches 'current_url'.
 
+=item pre_item_active
+
+INSTEAD of the "pre_item" string, use this string for active items
+
 =item post_active_item
 
 An additional string to append to each active item, before post_item.
@@ -263,17 +264,7 @@ sub link_list {
     }
     my %format = (exists $args{format}
 	? %{$args{format}}
-	: (
-	    pre_item=>$args{pre_item},
-	    post_item=>$args{post_item},
-	    pre_active_item=>$args{pre_active_item},
-	    post_active_item=>$args{post_active_item},
-	    pre_current_parent=>$args{pre_current_parent},
-	    post_current_parent=>$args{post_current_parent},
-	    pre_desc=>$args{pre_desc},
-	    post_desc=>$args{post_desc},
-	    item_sep=>$args{item_sep},
-	));
+	: make_default_format(%args));
     # correct the current_url
     $args{current_url} = make_canonical($args{current_url});
     my %current_parents = extract_current_parents(%args);
@@ -443,6 +434,10 @@ An additional string to put in front of each "active" item, after pre_item.
 The "active" item is the link which matches 'current_url'.
 (default: <em>)
 
+=item pre_item_active
+
+INSTEAD of the "pre_item" string, use this string for active items
+
 =item post_active_item
 
 An additional string to append to each active item, before post_item.
@@ -453,6 +448,11 @@ An additional string to append to each active item, before post_item.
 An additional string to put in front of a link which is a parent
 of the 'current_url' link, after pre_item.
 
+=item pre_item_current_parent
+
+INSTEAD of the "pre_item" string, use this for links which are parents
+of the 'current_url' link.
+
 =item post_current_parent
 
 An additional string to append to a link which is a parent
@@ -1022,10 +1022,10 @@ sub nav_tree {
 
     # correct the current_url
     $args{current_url} = make_canonical($args{current_url});
-    my $current_is_index = ($args{current_url} =~ m#/$#);
+    my $current_is_index = ($args{current_url} =~ m!/$!o);
     my %current_parents = extract_current_parents(%args);
 
-    # set the end depth if isn't already set
+    # set the end depth if is not already set
     # if this is an index-page, then make the depth its depth + 1
     # if this is a content-page, make the depth its depth
     my $current_url_depth = path_depth($args{current_url});
@@ -1103,7 +1103,7 @@ See L</link_list> for the formatting options.
 
 The label of the required link.  If there is no label,
 this uses the base-name of the last part of the link,
-capitalizing it and replacing underscores with spaces.
+capitalizing it and replacing underscores and dashes with spaces.
 
 =item this_link
 
@@ -1157,23 +1157,24 @@ sub make_item {
     if (!$label)
     {
 	$label = $link if !$label;
-	if ($link =~ /(\w+)\.\w+$/) # file
+	if ($link =~ /([-\w]+)\.\w+$/o) # file
 	{
 	    $label = $1;
 	}
-	elsif ($link =~ /(\w+)\/?$/) # dir
+	elsif ($link =~ /([-\w]+)\/?$/o) # dir
 	{
 	    $label = $1;
 	}
 	else # give up
 	{
 	    $label = $link;
-	    $label =~ s#/# :: #g;
+	    $label =~ s#/# :: #go;
 	}
 	
 	# prettify
-	$label =~ s#_# #g;
-	$label =~ s/(\w+)/\u\L$1/g;
+	$label =~ s#_# #go;
+	$label =~ s#-# #go;
+	$label =~ s/(\b[a-z][-\w]+)/\u\L$1/go;
     }
     # if we are hiding the extensions of files
     # we need to display an extensionless link
@@ -1181,7 +1182,7 @@ sub make_item {
     my $display_link = $link;
     if ($args{hide_ext})
     {
-	if ($link =~ /(.*)\.\w+$/) # file
+	if ($link =~ /(.*)\.[-\w]+$/o) # file
 	{
 	    $display_link = $1;
 	}
@@ -1199,7 +1200,7 @@ sub make_item {
     if (link_is_active(this_link=>$link,
 	current_url=>$args{current_url}))
     {
-	$item = join('', $format{pre_item},
+	$item = join('', $format{pre_item_active},
 		     $format{pre_active_item},
 		     $label,
 		     $format{post_active_item},
@@ -1216,7 +1217,7 @@ sub make_item {
 	and exists $args{current_parents}->{$link}
 	and $args{current_parents}->{$link})
     {
-	$item = join('', $format{pre_item},
+	$item = join('', $format{pre_item_current_parent},
 		     $format{pre_current_parent},
 		     '<a href="', $prefix_url, $display_link, '">',
 		     $label, '</a>',
@@ -1249,17 +1250,17 @@ sub make_canonical {
     my $url = shift;
 
     return $url if (!$url);
-    if ($url =~ m#^(/)index\.\w+$#)
+    if ($url =~ m{^/index\.\w+$}o)
+    {
+	$url = '/';
+    }
+    elsif ($url =~ m{^(.*/)index\.\w+$}o)
     {
 	$url = $1;
     }
-    elsif ($url =~ m#^(.*/)index\.\w+$#)
+    elsif ($url =~ m{/[-\w]+$}o) # no dots; a directory
     {
-	$url = $1;
-    }
-    elsif ($url =~ m#/\w+$#) # no dots; a directory
-    {
-	$url .= '/'; # add the slash
+	$url = join('', $url, '/'); # add the slash
     }
     return $url;
 } # make_canonical
@@ -1279,13 +1280,13 @@ sub get_index_path {
 
     return $url if (!$url);
     $url = make_canonical($url);
-    if ($url =~ m#^(.*)/\w+\.\w+$#)
+    if ($url =~ m{^(.*)/[-\w]+\.\w+$}o)
     {
 	$url = $1;
     }
-    elsif ($url ne '/')
+    elsif ($url ne '/' and $url =~ m{/$}o)
     {
-	$url =~ s#/$##;
+	chop $url;
     }
     return $url;
 } # get_index_path
@@ -1303,7 +1304,7 @@ sub get_index_parent {
 
     return $url if (!$url);
     $url = get_index_path($url);
-    if ($url =~ m#^(.*)/\w+$#)
+    if ($url =~ m#^(.*)/[-\w]+$#o)
     {
 	$url = $1;
     }
@@ -1321,10 +1322,11 @@ sub path_depth {
     my $url = shift;
 
     return 0 if ($url eq '/'); # root is zero
-    $url =~ s#/$##; # remove trailing /
-    $url =~ s#^/##; # remove leading /
-    my @url = split('/', $url);
-    return scalar @url;
+    if ($url =~ m!/$!o) # remove trailing /
+    {
+	chop $url;
+    }
+    return scalar ($url =~ tr!/!/!);
 } # path_depth
  
 =head2 link_is_active
@@ -1342,13 +1344,12 @@ sub link_is_active {
 		current_url=>'',
 		@_
 	       );
-    my $link = make_canonical($args{this_link});
-    my $current_url = $args{current_url};
-
     # if there is no current link, is not active.
-    return 0 if (!$current_url);
+    return 0 if (!$args{current_url});
 
-    return 1 if ($link eq $current_url);
+    my $link = make_canonical($args{this_link});
+
+    return 1 if ($link eq $args{current_url});
     return 0;
 
 } # link_is_active
@@ -1523,19 +1524,22 @@ sub extract_current_parents {
     );
     
     my %paths = ();
-    my $current_url = $args{current_url};
-    my @path_split = split('/', $current_url);
-    pop @path_split; # remove the current url
-    while (@path_split)
+    if ($args{current_url})
     {
-	# these paths are index-pages. should end in '/'
-	my $newpath = join('/', @path_split, '');
-	$paths{$newpath} = 1;
-	pop @path_split;
-    }
-    if ($args{exclude_root_parent})
-    {
-	delete $paths{"/"};
+	my $current_url = $args{current_url};
+	my @path_split = split('/', $current_url);
+	pop @path_split; # remove the current url
+	    while (@path_split)
+	    {
+		# these paths are index-pages. should end in '/'
+		my $newpath = join('/', @path_split, '');
+		$paths{$newpath} = 1;
+		pop @path_split;
+	    }
+	if ($args{exclude_root_parent})
+	{
+	    delete $paths{"/"};
+	}
     }
 
     return %paths;
@@ -1582,7 +1586,7 @@ sub build_lol {
 	my $path = $paths_ref->[0];
 	my $can_path = make_canonical($path);
 	my $path_depth = path_depth($can_path);
-	my $path_is_index = ($can_path =~ m#/$#);
+	my $path_is_index = ($can_path =~ m#/$#o);
 	if ($path_depth == $depth)
 	{
 	    shift @{$paths_ref}; # use this path
@@ -1670,19 +1674,17 @@ sub filter_out_paths {
     my $paths_ref = $args{paths};
     my $hide = $args{hide};
     my $nohide = $args{nohide};
-    my $current_url_depth = path_depth($args{current_url});
-    my $current_url_is_index = ($args{current_url} =~ m#/$#);
-    # the current-url dir is the current url without the filename
-    my $current_index_path = get_index_path($args{current_url});
-    my $current_index_path_depth = path_depth($current_index_path);
-    my $current_index_parent = get_index_parent($args{current_url});
 
-    my @wantedpaths = ();
+    my %canon_paths = ();
+    my @wantedpaths1 = ();
+    my %path_depth = ();
+
+    # filter out common things
+    # remember canonical paths and path depths
     foreach my $path (@{$paths_ref})
     {
 	my $can_path = make_canonical($path);
 	my $path_depth = path_depth($can_path);
-	my $path_is_index = ($can_path =~ m#/$#);
 	if ($hide and $nohide
 	    and not($path =~ /$nohide/)
 	    and $path =~ /$hide/)
@@ -1702,74 +1704,114 @@ sub filter_out_paths {
 	{
 	    # skip this one
 	}
-	# a breadcrumb-navbar shows the parent, self,
-	# and the children of dirs or siblings of non-dirs
-	elsif ($args{navbar_type} eq 'breadcrumb'
-	    and $args{current_url}
-	    and !(
-	     ($path_depth <= $current_url_depth
-	      and $args{current_url} =~ /^$path/)
-	     or (
-		 $path eq $args{current_url}
-		)
-	     or (
-		 $current_url_is_index
-		 and $path_depth >= $current_url_depth
-		 and $path =~ /^$current_index_path\//
-		)
-	     or (
-		 !$current_url_is_index
-		 and $path_depth >= $current_url_depth
-		 and $path =~ /^$current_index_parent\//
-		)
-	    )
-	   )
+	else
 	{
-	    # skip this one
+	    $path_depth{$path} = $path_depth;
+	    $canon_paths{$path} = $can_path;
+	    push @wantedpaths1, $path;
 	}
-	# a navbar shows the parent, the children
-	# and the current level
-	# and the top level (if we are starting at $top_level)
-	# and the siblings of one's parent if one is a contents-page
-	# or siblings of oneself if one is an index-page
-	elsif (($args{navbar_type}
-	    or $args{do_navbar}) # backwards compatibility
-	    and $args{current_url}
-	    and !(
-	     ($path_depth <= $current_url_depth
-	      and $args{current_url} =~ /^$path/)
-	     or (
-		 $path eq $args{current_url}
-		)
-	     or (
-		 $path_depth >= $current_url_depth 
-		 and $path =~ /^$current_index_path\//
-		)
-	     or (
-		 $args{start_depth} == $args{top_level}
-		 and $path_depth == $args{start_depth}
-		)
-	     or (
-		 !$current_url_is_index
-		 and $path_depth == $current_url_depth - 1
-		 and $path =~ /^$current_index_parent\//
-		)
-	     or (
-		 $current_url_is_index
-		 and $path_depth == $current_url_depth
-		 and $path =~ /^$current_index_parent\//
-		)
-	    )
-	   )
+    }
+
+    my @wantedpaths = ();
+    if ($args{current_url})
+    {
+	my $current_url = $args{current_url};
+	my $current_url_depth = path_depth($args{current_url});
+	my $current_url_is_index = ($args{current_url} =~ m{/$}o);
+
+	my $parent = make_canonical($current_url_is_index
+		      ? get_index_parent($args{current_url})
+		      : get_index_path($args{current_url})
+		     );
+	my $parent_depth = path_depth($parent);
+	my $grandparent = ($parent_depth == 1
+			   ? '/'
+			   : make_canonical(get_index_parent($parent)));
+	my $greatgrandparent = ($parent_depth <= 1
+				? ''
+				: ($parent_depth == 2
+				   ? '/'
+				   : make_canonical(get_index_parent($grandparent))
+				  )
+			       );
+	my $current_index_path = get_index_path($args{current_url});
+	my $current_index_parent = get_index_parent($args{current_url});
+
+	if ($args{navbar_type} eq 'breadcrumb')
 	{
-	    # skip this one
+	    foreach my $path (@wantedpaths1)
+	    {
+		my $pd = $path_depth{$path};
+		# a breadcrumb-navbar shows the parent, self,
+		# and the children the parent
+		if ($pd <= $current_url_depth
+		    and $args{current_url} =~ /^$path/)
+		{
+		    push @wantedpaths, $path;
+		}
+		elsif ($path eq $args{current_url})
+		{
+		    push @wantedpaths, $path;
+		}
+		elsif ($pd >= $current_url_depth
+		    and $path =~ m{^${current_url}})
+		{
+		    push @wantedpaths, $path;
+		}
+		elsif ($parent
+		       and $pd >= $current_url_depth
+		       and $path =~ m{^$parent})
+		{
+		    push @wantedpaths, $path;
+		}
+	    }
+	}
+	elsif ($args{navbar_type} or $args{do_navbar})
+	{
+	    # Rules for navbars:
+	    # * if I am a leaf node, see my (great)uncles and siblings
+	    # * if have children, use the same data as my parent,
+	    #   plus my immediate children
+	    foreach my $path (@wantedpaths1)
+	    {
+		my $pd = $path_depth{$path};
+		if ($pd > $current_url_depth + 1)
+		{
+		    next;
+		}
+		if ($pd == $current_url_depth + 1
+		    and $path =~ m{^${current_url}})
+		{
+		    push @wantedpaths, $path;
+		}
+		elsif ($pd == $current_url_depth
+		       and $path =~ m{^${parent}})
+		{
+		    push @wantedpaths, $path;
+		}
+		elsif ($grandparent
+		       and $pd == $parent_depth
+		       and $path =~ m{^$grandparent})
+		{
+		    push @wantedpaths, $path;
+		}
+		elsif ($greatgrandparent
+		    and $pd == $parent_depth - 1
+		    and $path =~ m{^$greatgrandparent})
+		{
+		    push @wantedpaths, $path;
+		}
+	    }
 	}
 	else
 	{
-	    # keep this path
-	    push @wantedpaths, $path;
+	    push @wantedpaths, @wantedpaths1;
 	}
     }
+    else
+    {
+	push @wantedpaths, @wantedpaths1;
+    }
     return @wantedpaths;
 } # filter_out_paths
 
@@ -1812,6 +1854,13 @@ sub make_default_format {
 			  tree_sep=>$args{tree_sep},
 			  tree_head=>$args{links_head},
 			  tree_foot=>$args{links_foot},
+			  pre_item_active=>($args{pre_item_active}
+					    ? $args{pre_item_active}
+					    : $args{pre_item}),
+			  pre_item_current_parent=>
+			  ($args{pre_item_current_parent}
+			   ? $args{pre_item_current_parent}
+			   : $args{pre_item}),
 			 );
     return %default_format;
 } # make_default_format
@@ -1836,6 +1885,8 @@ sub make_extra_formats {
 		last_subtree_foot=>"\n</ul>",
 		pre_item=>'<li>',
 		post_item=>'</li>',
+		pre_item_active=>'<li>',
+		pre_item_current_parent=>'<li>',
 		pre_active_item=>'<em>',
 		post_active_item=>'</em>',
 		pre_current_parent=>'',
